from django.http import HttpRequest

from django2_alexa.interfaces.request.base import BaseRequest
from django2_alexa.utils.alexa.locales import Locale


class LaunchRequest(BaseRequest):
    def __init__(self, request: HttpRequest):
        super().__init__(request)
        self.type = "LaunchRequest"
        self.request_id = self.body["requestId"]    # type: str
        self.timestamp = self.body["timestamp"]     # type: str
        self.locale = Locale(self.body["locale"])   # type: Locale



import json

from django.http import HttpRequest


class BaseRequest:
    def __init__(self, request: HttpRequest):
        self.body = json.loads(request.body.decode())   # type: dict
        raise Exception(request.body.decode())



from enum import Enum


class SpeechType(Enum):
    PLAIN_TEXT = "PlainText"
    SSML = "SSML"


class PlayBehavior(Enum):
    ENQUEUE = "ENQUEUE"
    REPLACE_ALL = "REPLACE_ALL"
    REPLACE_ENQUEUED = "REPLACE_ENQUEUED"


class OutputSpeech:
    def __init__(self, content: str, sp_type=SpeechType.PLAIN_TEXT, behavior=PlayBehavior.ENQUEUE):
        self.type = sp_type
        self.content = content
        self.behavior = behavior

    def to_dict(self):
        d = {
            'type': self.type.value,
            'playBehavior': self.behavior.value
        }
        if self.type == SpeechType.PLAIN_TEXT:
            d['text'] = self.content
        else:
            d['ssml'] = self.content
        return d



# Source: https://github.com/johnwheeler/flask-ask/blob/master/flask_ask/verifier.py
# License: Apache-2.0 https://github.com/johnwheeler/flask-ask/blob/master/LICENSE.txt
# Edited by Tim Woocker (Nov 2018)

import base64
import json
import posixpath
from datetime import datetime
from urllib.parse import urlparse
from urllib.request import urlopen

import aniso8601
from OpenSSL import crypto
from django.conf import settings
from django.http import HttpRequest

from . import VerificationError


def is_valid_request(request: HttpRequest) -> bool:
    url = request.META.get("SignatureCertChainUrl")
    signature = request.META.get("Signature")
    body = json.loads(request.body.decode())
    if not _valid_certificate_url(url):
        return False
    cert_data = urlopen(url).read()
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert_data)
    if not _valid_certificate(cert):
        return False
    if not verify_signature(cert, signature, request.body):
        return False
    if getattr(settings, "DEBUG", False):
        timestamp = _parse_timestamp(body.get('request', {}).get('timestamp'))
        if not verify_timestamp(timestamp):
            return False

    # TODO: Check application id on request
    # try:
    #     application_id = body['session']['application']['applicationId']
    # except KeyError:
    #     application_id = body['context']['System']['application']['applicationId']


def verify_signature(cert, signature, signed_data):
    try:
        signature = base64.b64decode(signature)
        crypto.verify(cert, signature, signed_data, 'sha1')
    except crypto.Error:
        return False
    return True


def verify_timestamp(timestamp) -> bool:
    dt = datetime.utcnow() - timestamp.replace(tzinfo=None)
    return not(abs(dt.total_seconds()) > 150)


def verify_application_id(candidate, records):
    if candidate not in records:
        raise VerificationError("Application ID verification failed")


def _valid_certificate_url(cert_url):
    parsed_url = urlparse(cert_url)
    if parsed_url.scheme == 'https':
        if parsed_url.hostname == "s3.amazonaws.com":
            if posixpath.normpath(parsed_url.path).startswith("/echo.api/"):
                return True
    return False


def _valid_certificate(cert):
    not_after = cert.get_notAfter().decode('utf-8')
    not_after = datetime.strptime(not_after, '%Y%m%d%H%M%SZ')
    if datetime.utcnow() >= not_after:
        return False
    found = False
    for i in range(0, cert.get_extension_count()):
        extension = cert.get_extension(i)
        short_name = extension.get_short_name().decode('utf-8')
        value = str(extension)
        if 'subjectAltName' == short_name and 'DNS:echo-api.amazon.com' == value:
            found = True
            break
    if not found:
        return False
    return True


def _parse_timestamp(timestamp):
    """
    Parse a given timestamp value, raising ValueError if None or Flasey
    """
    if timestamp:
        try:
            return aniso8601.parse_datetime(timestamp)
        except AttributeError:
            # raised by aniso8601 if raw_timestamp is not valid string
            # in ISO8601 format
            try:
                return datetime.utcfromtimestamp(timestamp)
            except:
                # relax the timestamp a bit in case it was sent in millis
                return datetime.utcfromtimestamp(timestamp / 1000)

    raise ValueError('Invalid timestamp value! Cannot parse from either ISO8601 string or UTC timestamp.')



from django.conf import settings
from django.http import HttpResponseServerError

from django2_alexa.interfaces.request.alexa import LaunchRequest
from django2_alexa.utils.s3_verification import is_valid_request


def launch(func):
    def wrapper(request, *args, **kwargs):
        if getattr(settings, "ALEXA_VERIFY_CONN", False) and not is_valid_request(request):
            # TODO: Troubleshooting part in docs
            return HttpResponseServerError("Amazon Server verification failed.")
        lr = LaunchRequest(request)
        return func(request, lr, *args, **kwargs)
    return wrapper


def intent(name: str):
    def inner(func):
        def wrapper(request, *args, **kwargs):
            if getattr(settings, "ALEXA_VERIFY_CONN", False) and not is_valid_request(request):
                return HttpResponseServerError("Amazon Server verification failed.")
            return func(request, *args, **kwargs)
        return wrapper
    return inner



import os

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setuptools.setup(
    name="django2-alexa",
    version="0.1.2",
    author="Tim Woocker & Malte Mosler",
    author_email="tim.woocker@googlemail.com",
    description="Django app for easily creating Amazon Alexa Skills",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/timwoocker/django-alexa",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import enum


class Locale(enum.Enum):
    ENGLISH_AU = "en-AU"
    ENGLISH_CA = "en-CA"
    ENGLISH_GB = "en-GB"
    ENGLISH_IN = "en-IN"
    ENGLISH_US = "en-US"
    ENGLISH_ES = "en-ES"
    GERMAN = "de-DE"
    SPANISH_ES = "es-ES"
    SPANISH_MX = "es-MX"
    FRENCH_CA = "fr-CA"
    FRENCH_FR = "fr-FR"
    ITALIAN = "it-IT"
    JAPANESE = "ja-JP"



class Card:
    def to_dict(self):
        pass


class SimpleCard(Card):
    def __init__(self, title: str=None, content: str=None):
        self.type = "Simple"
        self.title = title
        self.content = content

    def to_dict(self):
        d = {
            'type': self.type,
            'title': self.title,
            'content': self.content
        }
        return d


class StandardCard(Card):
    def __init__(self, title: str=None, text: str=None, small_image_url: str=None, large_item_url: str=None):
        self.type = "Standard"
        self.title = title
        self.text = text
        self.small_image_url = small_image_url
        self.large_item_url = large_item_url

    def to_dict(self):
        d = {
            'type': self.type,
            'title': self.title,
            'text': self.text
        }
        if self.small_image_url or self.large_item_url:
            d['image'] = {}
            if self.small_image_url:
                d['image']['smallImageUrl'] = self.small_image_url
            if self.large_item_url:
                d['image']['large_item_url'] = self.large_item_url

        return d


class LinkAccountCard(Card):
    def __init__(self):
        self.type = "LinkAccount"

    def to_dict(self):
        d = {
            'type': self.type
        }
        return d


class AskForPermissionsConsentCard(Card):
    def __init__(self, title: str=None, content: str=None, text: str=None):
        self.type = "AskForPermissionsConsent"
        self.title = title
        self.content = content
        self.text = text

    def to_dict(self):
        d = {
            'type': self.type,
            'title': self.title,
            'content': self.content,
            'text': self.text
        }
        return d



import json

from django.http import HttpResponse

from django2_alexa.interfaces.response.output_speech import OutputSpeech
from django2_alexa.utils.alexa.cards import Card


class Response(HttpResponse):
    def __init__(self, output_speech: OutputSpeech = None, card: Card = None, reprompt: OutputSpeech = None,
                 should_end_session=True, directives: [object] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.output_speech = output_speech
        self.card = card
        self.reprompt = reprompt
        self.should_session_end = should_end_session
        # TODO: directives

        self["Content-Type"] = "application/json;charset=UTF-8"

    def serialize(self):
        d = {
            'version': "1.0",
            # TODO: 'sessionAttributes': None,
            'response': {
                'outputSpeech': self.output_speech.to_dict(),
                'card': self.card.to_dict(),
                'reprompt': self.reprompt.to_dict(),
                'shouldEndSession': self.should_session_end,
                # TODO: directives
            }
        }
        return self.serialize_headers() + b'\r\n\r\n' + json.dumps(d)
